"""Provides a raw console to test module and demonstrate usage."""
import argparse
import asyncio
import logging

import insteonplm

__all__ = ('console', 'monitor')


@asyncio.coroutine
def console(loop, log, devicelist):
    """Connect to receiver and show events as they occur.

    Pulls the following arguments from the command line (not method arguments):

    :param device:
        Unix device where the PLM is attached
    :param verbose:
        Show debug logging.
    """
    parser = argparse.ArgumentParser(description=console.__doc__)
    parser.add_argument('--device', default='/dev/ttyUSB0', help='Device')
    parser.add_argument('--verbose', '-v', action='count')

    args = parser.parse_args()

    if args.verbose:
        level = logging.DEBUG
    else:
        level = logging.INFO

    logging.basicConfig(level=level)

    device = args.device

    log.info('Connecting to Insteon PLM at %s', device)

    conn = yield from insteonplm.Connection.create(
        device=device, loop=loop, userdefined=devicelist)

    def async_insteonplm_light_callback(device):
        """Log that our new device callback worked."""
        log.warn('New Device: %s', device)

    criteria = {}
    conn.protocol.add_device_callback(async_insteonplm_light_callback, criteria)

    plm = conn.protocol

    # yield from asyncio.sleep(5, loop=loop)
    #  Successfully turns off the light in my computer room (yay)
    #  conn.protocol._send_raw(binascii.unhexlify('02624095e6001300'))
    #
    # conn.protocol._send_raw(binascii.unhexlify('02624095e6000300'))
    # conn.protocol.product_data_request('15c3ab')
    # yield from asyncio.sleep(10, loop=loop)

    yield from asyncio.sleep(20, loop=loop)

    if 1 == 0:
        # conn.protocol.send_insteon_extended('4095e6', '2e', '00', '0000000000000000000000000000')
        # INFO:insteonplm.protocol:INSTEON extended 40.95.E6->39.55.37: cmd1:2e
        # cmd2:00 flags:15 data:b'0101000020201cff1f0001000000'
        # d2: 01
        #
        conn.protocol.extended_status_request('4095e6')
        yield from asyncio.sleep(5, loop=loop)

        conn.protocol.update_setlevel('4095e6', 127)
        conn.protocol.update_ramprate('4095e6', 27)
        yield from asyncio.sleep(5, loop=loop)

        conn.protocol.extended_status_request('4095e6')
        yield from asyncio.sleep(5, loop=loop)

    if 1 == 0:
        conn.protocol.turn_on('4095e6', ramprate=2)
        yield from asyncio.sleep(5, loop=loop)

    if 1 == 0:
        conn.protocol.text_string_request('4095e6')

    if 1 == 0:
        conn.protocol.product_data_request('395fa4')

    if 1 == 0:
        conn.protocol.product_data_request('4095e6')
        conn.protocol.product_data_request('395fa4')
        conn.protocol.text_string_request('4095e6')
        yield from asyncio.sleep(5, loop=loop)

    # conn.protocol._send_raw(binascii.unhexlify('02624095e6150300000000000000ffff000000000000'))
    # yield from asyncio.sleep(5, loop=loop)

    if 1 == 0:
        print('-- ')
        conn.protocol.relay_request('395fa4')
        conn.protocol.relay_request('395ecb')
        print('-- ')
        conn.protocol.turn_off('395fa4')
        conn.protocol.turn_off('395ecb')
        print('-- ')
        conn.protocol.turn_on('395fa4', 1)
        yield from asyncio.sleep(5, loop=loop)
        print('-- ')
        conn.protocol.turn_on('395ecb', 1)
        yield from asyncio.sleep(5, loop=loop)

    if 1 == 0:
        yield from asyncio.sleep(5, loop=loop)
        conn.protocol.dump_all_link_database()
        yield from asyncio.sleep(5, loop=loop)

    if 1 == 0:
        addr = '395fa4'
        addr = '424356'
        addr = '395ecb'
        log.info('Are you ready to rumble?')
        yield from asyncio.sleep(2, loop=loop)
        plm.status_request(addr)
        yield from asyncio.sleep(5, loop=loop)
        plm.turn_on(addr)
        yield from asyncio.sleep(5, loop=loop)
        plm.status_request(addr)
        yield from asyncio.sleep(5, loop=loop)
        plm.status_request(addr)

    if 1 == 1:
        for addr in ['395ecb', '395fa4']:
            log.info('Are you ready to rumble?')
            yield from asyncio.sleep(2, loop=loop)
            log.info('Relay Status')
            plm.status_request(addr)
            yield from asyncio.sleep(2, loop=loop)
            log.info('Sensor Status')
            plm.status_request(addr, '01')
            yield from asyncio.sleep(10, loop=loop)

def monitor():
    """Wrapper to call console with a loop."""
    devicelist = (
        {
            "address": "3c4fc5",
            "cat": 0x05,
            "subcat": 0x0b,
            "firmware": 0x00
        },
        {
            "address": "43af9b",
            "cat": 0x02,
            "subcat": 0x1a,
            "firmware": 0x00
        }
    )
    log = logging.getLogger(__name__)
    loop = asyncio.get_event_loop()
    asyncio.async(console(loop, log, devicelist))
    loop.run_forever()



"""Module containing the connection wrapper for the PLM interface."""
import asyncio
import logging
import serial
import serial.aio

from .protocol import PLM

__all__ = ('Connection')

# pylint: disable=invalid-name,no-member
try:
    ensure_future = asyncio.ensure_future
except AttributeError:
    ensure_future = asyncio.async


class Connection:
    """Handler to maintain the Powerline device connection."""

    def __init__(self):
        """Instantiate the Connection object."""
        self.log = logging.getLogger(__name__)

    @classmethod
    @asyncio.coroutine
    def create(cls, device='/dev/ttyUSB0',
               auto_reconnect=True, loop=None, protocol_class=PLM, userdefined=()):
        """Initiate a connection to a specific device.

        Here is where we supply the device and callback callables we
        expect for this PLM class object.

        :param device:
            Unix device where the PLM is attached
        :param auto_reconnect:
            Should the Connection try to automatically reconnect if needed?
        :param loop:
            asyncio.loop for async operation

        :type device:
            str
        :type auto_reconnect:
            boolean
        :type loop:
            asyncio.loop
        :type update_callback:
            callable
        """
        conn = cls()

        conn.device = device
        conn._loop = loop or asyncio.get_event_loop()
        conn._retry_interval = 1
        conn._closed = False
        conn._closing = False
        conn._halted = False
        conn._auto_reconnect = auto_reconnect

        def connection_lost():
            """Function callback for Protocol when connection is lost."""
            if conn._auto_reconnect and not conn._closing:
                ensure_future(conn._reconnect(), loop=conn._loop)

        conn.protocol = protocol_class(
            connection_lost_callback=connection_lost, loop=conn._loop, userdefineddevices=userdefined)

        yield from conn._reconnect()

        return conn

    @property
    def transport(self):
        """Return pointer to the transport object.

        Use this property to obtain passthrough access to the underlying
        Protocol properties and methods.
        """
        return self.protocol.transport

    def _get_retry_interval(self):
        return self._retry_interval

    def _reset_retry_interval(self):
        self._retry_interval = 1

    def _increase_retry_interval(self):
        self._retry_interval = min(300, 1.5 * self._retry_interval)

    @asyncio.coroutine
    def _reconnect(self):
        while True:
            try:
                if self._halted:
                    yield from asyncio.sleep(2, loop=self._loop)
                else:
                    self.log.info('Connecting to PLM on %s', self.device)
                    yield from serial.aio.create_serial_connection(
                        self._loop, lambda: self.protocol,
                        self.device, baudrate=19200)
                    self._reset_retry_interval()
                    return

            except OSError:
                self._increase_retry_interval()
                interval = self._get_retry_interval()
                self.log.warning('Connecting failed, retry in %i seconds: %s',
                                 interval, self.device)
                yield from asyncio.sleep(interval, loop=self._loop)

    def close(self):
        """Close the PLM device connection and don't try to reconnect."""
        self.log.warning('Closing connection to PLM')
        self._closing = True
        if self.protocol.transport:
            self.protocol.transport.close()

    def halt(self):
        """Close the PLM device connection and wait for a resume() request."""
        self.log.warning('Halting connection to PLM')
        self._halted = True
        if self.protocol.transport:
            self.protocol.transport.close()

    def resume(self):
        """Resume the PLM device connection if we have been halted."""
        self.log.warning('Resuming connection to PLM')
        self._halted = False

    @property
    def dump_conndata(self):
        """Developer tool for debugging forensics."""
        attrs = vars(self)
        return ', '.join("%s: %s" % item for item in attrs.items())



"""Helper objects for maintaining PLM state and interfaces."""
import logging
import binascii

__all__ = ('Address', 'PLMProtocol')

class Address(object):
    """Datatype definition for INSTEON device address handling."""

    def __init__(self, addr):
        """Create an Address object."""
        self.log = logging.getLogger(__name__)
        self.addr = self.normalize(addr)

    def __repr__(self):
        return self.human

    def __str__(self):
        return self.hex

    def __eq__(self, other):
        return self.hex == other.hex

    def __ne__(self, other):
        return self.hex != other.hex

    def normalize(self, addr):
        """Take any format of address and turn it into a hex string."""
        if isinstance(addr, Address):
            return addr.hex
        if isinstance(addr, bytearray):
            return binascii.hexlify(addr).decode()
        if isinstance(addr, bytes):
            return binascii.hexlify(addr).decode()
        if isinstance(addr, str):
            addr.replace('.', '')
            addr = addr[0:6]
            return addr.lower()
        else:
            self.log.warning('Address class init with unknown type %s: %r',
                             type(addr), addr)
            return '000000'

    @property
    def human(self):
        """Emit the address in human-readible format (AA.BB.CC)."""
        addrstr = self.addr[0:2]+'.'+self.addr[2:4]+'.'+self.addr[4:6]
        return addrstr.upper()

    @property
    def hex(self):
        """Emit the address in bare hex format (aabbcc)."""
        return self.addr

    @property
    def bytes(self):
        r"""Emit the address in bytes format (b'\xaabbcc')."""
        return binascii.hexlify(self.addr)


class PLMCode(object):
    """Class to store PLM code definitions and attributes."""

    def __init__(self, code, name=None, size=None, rsize=None):
        """Create a new PLM code object."""
        self.code = code
        self.size = size
        self.rsize = rsize
        self.name = name


class PLMProtocol(object):
    """Class container to store PLMCode objects as a Protocol."""

    def __init__(self):
        """Create the Protocol object."""
        self.log = logging.getLogger(__name__)
        self._codelist = []
        self.add(0x50, name='INSTEON Standard Message Received', size=11)
        self.add(0x51, name='INSTEON Extended Message Received', size=25)
        self.add(0x52, name='X10 Message Received', size=4)
        self.add(0x53, name='ALL-Linking Completed', size=10)
        self.add(0x54, name='Button Event Report', size=3)
        self.add(0x55, name='User Reset Detected', size=2)
        self.add(0x56, name='ALL-Link CLeanup Failure Report', size=2)
        self.add(0x57, name='ALL-Link Record Response', size=10)
        self.add(0x58, name='ALL-Link Cleanup Status Report', size=3)
        self.add(0x60, name='Get IM Info', size=2, rsize=9)
        self.add(0x61, name='Send ALL-Link Command', size=5, rsize=6)
        self.add(0x62, name='INSTEON Fragmented Message', size=8, rsize=9)
        self.add(0x69, name='Get First ALL-Link Record', size=2)
        self.add(0x6a, name='Get Next ALL-Link Record', size=2)
        self.add(0x73, name='Get IM Configuration', size=2, rsize=6)


    def __len__(self):
        """Return the number of PLMCodes in the Protocol."""
        return len(self._codelist)

    def __iter__(self):
        for x in self._codelist:
            yield x.code

    def add(self, code, name=None, size=None, rsize=None):
        """Add a new PLMCode to the Protocol."""
        self._codelist.append(PLMCode(code, name=name, size=size, rsize=rsize))

    def lookup(self, code, fullmessage=None):
        """Return the PLMCode from a byte and optional stream buffer."""
        for x in self._codelist:
            if x.code == code:
                if code == 0x62 and fullmessage:
                    x.name = 'INSTEON Fragmented Message'
                    x.size = 8
                    x.rsize = 9
                    if len(fullmessage) >= 6:
                        flags = fullmessage[5]
                        if flags == 0x00:
                            x.name = 'INSTEON Standard Message'
                        else:
                            x.name = 'INSTEON Extended Message'
                            x.size = 22
                            x.rsize = 23

                return x

class Message(object):
    """Unroll a raw message string into a class with attributes."""
    def __init__(self, rawmessage):
        self.log = logging.getLogger(__name__)
        self.code = rawmessage[1]
        self.rawmessage = rawmessage

        if self.code == 0x50 or self.code == 0x51:
            # INSTEON Standard and Extended Message
            self.address = Address(rawmessage[2:5])
            self.target = Address(rawmessage[5:8])
            self.flagsval = rawmessage[8]
            self.cmd1 = rawmessage[9]
            self.cmd2 = rawmessage[10]
            self.flags = self.decode_flags(self.flagsval)
            self.userdata = rawmessage[11:25]

        elif self.code == 0x53:
            # ALL-Linking Complete
            self.linkcode = rawmessage[2]
            self.group = rawmessage[3]
            self.address = Address(rawmessage[4:7])
            self.category = rawmessage[7]
            self.subcategory = rawmessage[8]
            self.firmware = rawmessage[9]

        elif self.code == 0x54:
            events = {0x02: 'SET button tapped',
                      0x03: 'SET button press and hold',
                      0x04: 'SET button released',
                      0x12: 'Button 2 tapped',
                      0x13: 'Button 2 press and hold',
                      0x14: 'Button 2 released',
                      0x22: 'Button 3 tapped',
                      0x23: 'Button 3 press and hold',
                      0x24: 'Button 3 released'}

            self.event = rawmessage[2]
            self.description = events.get(self.event, None)

        elif self.code == 0x57:
            # ALL-Link Record Response
            self.flagsval = rawmessage[2]
            self.group = rawmessage[3]
            self.address = Address(rawmessage[4:7])
            self.linkdata1 = rawmessage[7]
            self.linkdata2 = rawmessage[8]
            self.linkdata3 = rawmessage[9]

        elif self.code == 0x60:
            self.address = Address(rawmessage[2:5])
            self.category = rawmessage[5]
            self.subcategory = rawmessage[6]
            self.firmware = rawmessage[7]

        elif self.code == 0x62:
            # 0262395fa4001900
            self.address = Address(rawmessage[2:5])
            self.flagsval = rawmessage[5]

        elif self.code == 0x73:
            self.flagsval = rawmessage[2]
            self.spare1 = rawmessage[3]
            self.spare2 = rawmessage[4]

    def __repr__(self):
        attrs = vars(self)
        return ', '.join("%s: %r" % item for item in attrs.items())

    def decode_flags(self, flags):
        """Turn INSTEON message flags into a dict."""
        retval = {}
        if flags is not None:
            retval['broadcast'] = (flags & 128) > 0
            retval['group'] = (flags & 64) > 0
            retval['ack'] = (flags & 32) > 0
            retval['extended'] = (flags & 16) > 0
            retval['hops'] = (flags & 12 >> 2)
            retval['maxhops'] = (flags & 3)
        return retval



"""Module to maintain PLM state information and network interface."""
import asyncio
import logging
import binascii

from .ipdb import IPDB
from .plm import Address, PLMProtocol, Message

__all__ = ('PLM', 'ALDB')

PP = PLMProtocol()


class ALDB(object):
    """Class holds and maintains the ALL-Link Database from the PLM device."""

    ipdb = IPDB()

    def __init__(self):
        """Instantiate the ALL-Link Database object."""
        self.log = logging.getLogger(__name__)
        self.state = 'empty'
        self._devices = {}
        self._cb_new_device = []
        self._overrides = {}

    def __len__(self):
        """Return the number of devices in the ALDB."""
        return len(self._devices)

    def __iter__(self):
        """Iterate through each ALDB device record."""
        for device in self._devices:
            yield device

    def __getitem__(self, address):
        """Fetch a device from the ALDB."""
        if address in self._devices:
            return self._devices[address]
        raise KeyError

    def __setitem__(self, key, value):
        """Add or Update a device in the ALDB."""
        if 'cat' not in value or value['cat'] == 0:
            self.log.debug('Ignoring device setitem with no cat: %s', value)
            return

        if key in self._devices:
            if 'firmware' in value and value['firmware'] < 255:
                self._devices[key].update(value)
        else:
            productdata = self.ipdb[value['cat'], value['subcat']]
            value.update(productdata._asdict())
            address = Address(key)
            value['address_hex'] = address.hex
            value['address'] = address.human
            self._devices[key] = value

            self.log.info('New INSTEON Device %r: %s (%02x:%02x)',
                          Address(key), value['description'], value['cat'],
                          value['subcat'])

            self._apply_overrides(key)

            for callback, criteria in self._cb_new_device:
                if self._device_matches_criteria(value, criteria):
                    callback(value)

    def __repr__(self):
        """Human representation of a device from the ALDB."""
        attrs = vars(self)
        return ', '.join("%s: %r" % item for item in attrs.items())

    def add_device_callback(self, callback, criteria):
        """Register a callback to be invoked when a new device appears."""
        self.log.info('New callback %s with %s (%d items already in list)',
                      callback, criteria, len(self._devices.keys()))
        self._cb_new_device.append([callback, criteria])

        #
        # When a new device callback is added, we want to include all
        # previously-discovered devices as well as new devices.  So we
        # iterate through the existing device list and trigger the callback
        # for each of them
        #
        for device in self:
            value = self[device]
            if self._device_matches_criteria(value, criteria):
                self.log.info('retroactive callback device %s matching %s',
                              value['address'], criteria)
                callback(value)

    def add_override(self, addr, key, value):
        """Register an attribute override for a device."""
        address = Address(addr).hex
        self.log.info('New override for %s %s is %s', address, key, value)
        device_override = self._overrides.get(address, {})
        device_override[key] = value
        self._overrides[address] = device_override

        if address in self._devices:
            self._apply_overrides(address)

    def _apply_overrides(self, address):
        device_override = self._overrides.get(address, {})
        for key in device_override:
            oldval = self._devices[address].get(key, None)
            value = device_override[key]
            if value != oldval:
                self.log.debug('Override %s for %s: %s -> %s',
                               key, address, oldval, value)
                self._devices[address][key] = value

    def getattr(self, key, attr):
        """Return the requested attribute for device with address 'key'."""
        key = Address(key).hex
        if key in self._devices:
            return self._devices[key].get(attr, None)
        return None

    def setattr(self, key, attr, value):
        """Set supplied attribute on designated device in the ALDB."""
        key = Address(key).hex

        if key in self._devices:
            oldvalue = self._devices[key].get(attr, None)
            self._devices[key][attr] = value
            if value != oldvalue:
                self.log.info('Device %s.%s changed: %s->%s"',
                              key, attr, oldvalue, value)
                return True
            else:
                self.log.debug('Device %s.%s unchanged: %s->%s"',
                               key, attr, oldvalue, value)
                return False
        else:
            raise KeyError

    @staticmethod
    def _device_matches_criteria(device, criteria):
        match = True

        if 'address' in criteria:
            criteria['address'] = Address(criteria['address'])

        for key in criteria.keys():
            if key == 'capability':
                if criteria[key] not in device['capabilities']:
                    match = False
                    break
            elif key[0] != '_':
                if key not in device:
                    match = False
                    break
                elif criteria[key] != device[key]:
                    match = False
                    break
        return match


# pylint: disable=too-many-instance-attributes, too-many-public-methods
class PLM(asyncio.Protocol):
    """The Insteon PLM IP control protocol handler."""

    def __init__(self, loop=None, connection_lost_callback=None, userdefineddevices=()):
        """Protocol handler that handles all status and changes on PLM.

        This class is expected to be wrapped inside a Connection class object
        which will maintain the socket and handle auto-reconnects.

            :param connection_lost_callback:
                called when connection is lost to device (optional)
            :param loop:
                asyncio event loop (optional)

            :type: connection_lost_callback:
                callable
            :type loop:
                asyncio.loop
        """
        self._loop = loop

        self._connection_lost_callback = connection_lost_callback
        self._update_callbacks = []
        self._message_callbacks = []
        self._insteon_callbacks = []

        self._me = None
        self._buffer = bytearray()
        self._last_message = None
        self._last_command = None
        self._wait_for = {}
        self._recv_queue = []
        self._send_queue = []

        self.devices = ALDB()

        self.log = logging.getLogger(__name__)
        self.transport = None

        self._userdefineddevices = {}

        for userdevice in userdefineddevices:
            if 'cat' in userdevice and 'subcat' in userdevice and 'firmware' in userdevice:
                self._userdefineddevices[userdevice["address"]] = {
                    "cat": userdevice["cat"],
                    "subcat": userdevice["subcat"],
                    "firmware": userdevice["firmware"],
                    "status": "notfound"
                }
                self.devices[userdevice["address"]] = {'cat': userdevice["cat"],
                                            'subcat': userdevice["subcat"],
                                            'firmware': userdevice["firmware"]}

        self.add_message_callback(self._parse_insteon_standard, {'code': 0x50})
        self.add_message_callback(self._parse_insteon_extended, {'code': 0x51})
        self.add_message_callback(self._parse_all_link_completed, {'code': 0x53})
        self.add_message_callback(self._parse_button_event, {'code': 0x54})
        self.add_message_callback(self._parse_all_link_record, {'code': 0x57})
        self.add_message_callback(self._parse_get_plm_info, {'code': 0x60})
        self.add_message_callback(self._parse_get_plm_config, {'code': 0x73})

        self.add_insteon_callback(self._insteon_on, {'cmd1': 0x11})
        self.add_insteon_callback(self._insteon_on, {'cmd1': 0x12})
        self.add_insteon_callback(self._insteon_off, {'cmd1': 0x13})
        self.add_insteon_callback(self._insteon_off, {'cmd1': 0x14})
        self.add_insteon_callback(self._insteon_manual_change_stop, {'cmd1': 0x18})

    #
    # asyncio network functions
    #

    def connection_made(self, transport):
        """Called when asyncio.Protocol establishes the network connection."""
        self.log.info('Connection established to PLM')
        self.transport = transport

        # self.transport.set_write_buffer_limits(128)
        # limit = self.transport.get_write_buffer_size()
        # self.log.debug('Write buffer size is %d', limit)
        self.get_plm_info()
        self.load_all_link_database()

    def data_received(self, data):
        """Called when asyncio.Protocol detects received data from network."""
        self.log.debug('Received %d bytes from PLM: %s',
                       len(data), binascii.hexlify(data))

        self._buffer.extend(data)
        self._peel_messages_from_buffer()

        for message in self._recv_queue:
            self._process_message(message)
            self._recv_queue.remove(message)

    def connection_lost(self, exc):
        """Called when asyncio.Protocol loses the network connection."""
        if exc is None:
            self.log.warning('eof from modem?')
        else:
            self.log.warning('Lost connection to modem: %s', exc)

        self.transport = None

        if self._connection_lost_callback:
            self._connection_lost_callback()

    def _rsize(self, message):
        code = message[1]
        ppc = PP.lookup(code, fullmessage=message)

        if hasattr(ppc, 'rsize') and ppc.rsize:
            self.log.debug('Found a code 0x%x message which returns %d bytes',
                           code, ppc.rsize)
            return ppc.rsize
        else:
            self.log.debug('Unable to find an rsize for code 0x%x', code)
            return len(message) + 1

    def _timeout_reached(self):
        self.log.debug('_timeout_reached invoked')
        self._clear_wait()
        self._process_queue()

    def _clear_wait(self):
        self.log.debug('_clear_wait invoked')
        if '_thandle' in self._wait_for:
            self._wait_for['_thandle'].cancel()
        self._wait_for = {}

    def _schedule_wait(self, keys, timeout=1):
        if self._wait_for != {}:
            self.log.warning('Overwriting stale wait_for: %s', self._wait_for)
            self._clear_wait()

        self.log.debug('Waiting for %s timeout in %d seconds', keys, timeout)
        if timeout > 0:
            keys['_thandle'] = self._loop.call_later(timeout,
                                                     self._timeout_reached)

        self._wait_for = keys

    def _wait_for_last_command(self):
        sentmessage = self._last_command
        rsize = self._rsize(sentmessage)
        self.log.debug('Wait for ACK/NAK on sent: %s expecting rsize of %d',
                       binascii.hexlify(sentmessage), rsize)
        if self._buffer.find(sentmessage) == 0:
            if len(self._buffer) < rsize:
                self.log.debug('Waiting for all of message to arrive, %d/%d',
                               len(self._buffer), rsize)
                return

            code = self._buffer[1]
            message_length = len(sentmessage)
            response_length = rsize - message_length
            response = self._buffer[message_length:response_length]
            acknak = self._buffer[rsize-1]

            mla = self._buffer[:rsize-1]
            buffer = self._buffer[rsize:]

            data_in_ack = [0x19, 0x2e]
            msg = Message(mla)

            queue_ack = False

            if sentmessage != mla:
                self.log.debug('sent command and ACK response differ')
                queue_ack = True

            if hasattr(msg, 'cmd1') and msg.cmd1 in data_in_ack:
                self.log.debug('sent cmd1 is on queue_ack list')
                queue_ack = True

            if acknak == 0x06:
                if len(response) > 0 or queue_ack is True:
                    self.log.debug('Sent command %s OK with response %s',
                                  binascii.hexlify(sentmessage), response)
                    self._recv_queue.append(mla)
                else:
                    self.log.debug('Sent command %s OK',
                                   binascii.hexlify(sentmessage))

            else:
                if code == 0x6a:
                    self.log.info('ALL-Link database dump is complete')
                    self.devices.state = 'loaded'
                    for address in self.devices:
                        device = self.devices[address]
                        if 'cat' in device and device['cat'] > 0:
                            self.log.debug('I know the category for %s (0x%x)',
                                           address, device['cat'])
                        else:
                            self.product_data_request(device)
                    for userdevice in self._userdefineddevices:
                        if self._userdefineddevices[userdevice]["status"] == "notfound":
                            self.log.info("Failed to discover device %r.", userdevice)
                    self.poll_devices()
                else:
                    self.log.warning('Sent command %s UNsuccessful! (%02x)',
                                     binascii.hexlify(sentmessage), acknak)
            self._last_command = None
            self._buffer = buffer

    def _wait_for_recognized_message(self):
        code = self._buffer[1]

        for ppcode in PP:
            if ppcode == code or ppcode == bytes([code]):
                ppc = PP.lookup(code, fullmessage=self._buffer)

                self.log.debug('Found a code %02x message which is %d bytes',
                               code, ppc.size)

                if len(self._buffer) == ppc.size:
                    new_message = self._buffer[0:ppc.size]
                    self.log.debug('new message is: %s',
                                   binascii.hexlify(new_message))
                    self._recv_queue.append(new_message)
                    self._buffer = self._buffer[ppc.size:]
                else:
                    self.log.debug('Need more bytes to process message.')

    def _peel_messages_from_buffer(self):
        lastlooplen = 0
        worktodo = True

        while worktodo:
            if len(self._buffer) == 0:
                self.log.debug('Clean break!  There is no buffer left')
                worktodo = False
                break

            if len(self._buffer) < 2:
                worktodo = False
                break

            if self._buffer[0] != 2:
                self._buffer = self._buffer[1:]
                self.log.debug('Trimming leading buffer garbage')

            if len(self._buffer) == lastlooplen:
                # Buffer size did not change so we should wait for more data
                worktodo = False
                break

            lastlooplen = len(self._buffer)

            if self._buffer.find(2) < 0:
                self.log.debug('Buffer does not contain a 2, we should bail')
                worktodo = False
                break

            if self._last_command:
                self._wait_for_last_command()
            else:
                self._wait_for_recognized_message()

        self._process_queue()

    def _process_queue(self):
        self.log.debug('Send queue contains %d items', len(self._send_queue))
        if self._clear_to_send() is True:
            command, wait_for = self._send_queue[0]
            self._send_hex(command, wait_for=wait_for)
            self._send_queue.remove([command, wait_for])

    def _clear_to_send(self):
        if len(self._buffer) == 0:
            if len(self._send_queue) > 0:
                if self._last_command is None:
                    if self._wait_for == {}:
                        return True

    def _process_message(self, rawmessage):
        self.log.debug('Processing message: %s', binascii.hexlify(rawmessage))
        if rawmessage[0] != 2 or len(rawmessage) < 2:
            self.log.warning('process_message called with a malformed message')
            return

        if rawmessage == self._last_message:
            self.log.debug('ignoring duplicate message')
            return

        self._last_message = rawmessage

        msg = Message(rawmessage)

        # if hasattr(msg, 'target') and msg.target != self._me:
        #     self.log.info('Ignoring message that is not for me')
        #     return

        if self._message_matches_criteria(msg, self._wait_for):
            if '_callback' in self._wait_for:
                self._clear_wait()
                self._process_queue()
                return
            else:
                self._clear_wait()

        callbacked = False
        for callback, criteria in self._message_callbacks:
            if self._message_matches_criteria(msg, criteria):
                self.log.debug('message callback %s with criteria %s',
                               callback, criteria)
                self._loop.call_soon(callback, msg)
                callbacked = True

        if callbacked is False:
            ppc = PP.lookup(msg.code, fullmessage=rawmessage)
            if hasattr(ppc, 'name') and ppc.name:
                self.log.info('Unhandled event: %s (%s)', ppc.name,
                              binascii.hexlify(rawmessage))
            else:
                self.log.info('Unrecognized event: UNKNOWN (%s)',
                              binascii.hexlify(rawmessage))

        self._process_queue()

    def _message_matches_criteria(self, msg, criteria):
        match = True

        if criteria is None or criteria == {}:
            return True

        if 'address' in criteria:
            criteria['address'] = Address(criteria['address'])

        for key in criteria.keys():
            if key[0] != '_':
                mattr = getattr(msg, key, None)
                if mattr is None:
                    match = False
                    break
                elif criteria[key] != mattr:
                    match = False
                    break

        if match is True:
            if '_callback' in criteria:
                self.log.debug('Calback invoked from mmc criteria')
                criteria['_callback'](msg)

        return match

    def _parse_insteon_standard(self, msg):
        device = self.devices[msg.address.hex]

        self.log.info('INSTEON standard %r->%r: cmd1:%02x cmd2:%02x flags:%02x',
                      msg.address, msg.target,
                      msg.cmd1, msg.cmd2, msg.flagsval)
        self.log.debug('flags: %r', msg.flags)
        self.log.debug('device: %r', device)

        for callback, criteria in self._insteon_callbacks:
            if self._message_matches_criteria(msg, criteria):
                self.log.debug('insteon callback %s with criteria %s',
                               callback, criteria)
                self._loop.call_soon(callback, msg, device)

    def _parse_insteon_extended(self, msg):
        device = self.devices[msg.address.hex]

        self.log.info('INSTEON extended %r->%r: cmd1:%02x cmd2:%02x flags:%02x data:%s',
                      msg.address, msg.target, msg.cmd1, msg.cmd2, msg.flagsval,
                      binascii.hexlify(msg.userdata))
        self.log.debug('flags: %r', msg.flags)
        self.log.debug('device: %r', device)

        if msg.cmd1 == 0x03 and msg.cmd2 == 0x00:
            self._parse_product_data_response(msg.address, msg.userdata)

    def _parse_status_response(self, msg):
        onlevel = msg.cmd2

        self.log.info('INSTEON device status %r is at level %s',
                      msg.address, hex(onlevel))
        self.devices.setattr(msg.address, 'onlevel', onlevel)
        self._do_update_callback(msg)

    def _parse_sensor_response(self, msg):
        device = self.devices[msg.address.hex]

        sensorstate = msg.cmd2

        if device.get('model') == '2450':
            # Swap the values for a 2450 because it's opposite
            self.log.debug('Reversing sensorstate %s because 2450', sensorstate)
            if sensorstate:
                sensorstate = 0
            else:
                sensorstate = 1

        self.log.info('INSTEON sensor status %r is at level %s',
                      msg.address, hex(sensorstate))
        self.devices.setattr(msg.address, 'sensorstate', sensorstate)
        self._do_update_callback(msg)

    def _insteon_on(self, msg, device):
        self.log.info('INSTEON on event: %r, %r', msg, device)

        attribute = 'onlevel'

        if 'binary_sensor' in device.get('capabilities'):
            if msg.flags.get('group', False):
                #
                # If this is a group message from a sensor device, then we
                # treat it as sensorstate instead of onlevel.  This is vital
                # for I/O Linc 2450 devices, because it's the cleanest way
                # to differntiate between the sensor on/off notifications and
                # the relay on/off notifications that come in response to a
                # turn_on or turn_off command.  Those on/off messages are
                # direct (not group, not broadcast) and reflect the relay's
                # status and not the sensor's status.
                #
                attribute = 'sensorstate'

        if msg.cmd2 == 0x00:
            value = 255
        else:
            value = msg.cmd2

        if device.get('model') == '2477D':
            if msg.cmd2 == 0x00 or msg.cmd2 == 0x01:
                value = device.get('setlevel', 255)
                self.log.debug('ON report with no onlevel, using %02x', value)

        if self.devices.setattr(msg.address, attribute, value):
            self._do_update_callback(msg)

    def _insteon_off(self, msg, device):
        self.log.info('INSTEON off event: %r, %r', msg, device)

        attribute = 'onlevel'

        if 'binary_sensor' in device.get('capabilities'):
            if msg.flags.get('group', False):
                #
                # If this is a group message from a sensor device, then we
                # treat it as sensorstate instead of onlevel.  This is vital
                # for I/O Linc 2450 devices, because it's the cleanest way
                # to differntiate between the sensor on/off notifications and
                # the relay on/off notifications that come in response to a
                # turn_on or turn_off command.  Those on/off messages are
                # direct (not group, not broadcast) and reflect the relay's
                # status and not the sensor's status.
                #
                attribute = 'sensorstate'

        if self.devices.setattr(msg.address, attribute, 0):
            self._do_update_callback(msg)

    # pylint: disable=unused-argument
    def _insteon_manual_change_stop(self, msg, device):
        self.log.info('Light Stop Manual Change')
        self.status_request(msg.address)

    def _parse_extended_status_response(self, msg):
        device = self.devices[msg.address.hex]

        self.log.info('INSTEON extended device status %r', msg.address)
        if device.get('cat') == 0x01:
            self.devices.setattr(msg.address, 'ramprate', msg.userdata[6])
            self.devices.setattr(msg.address, 'setlevel', msg.userdata[7])

    def _do_update_callback(self, msg):
        for callback, criteria in self._update_callbacks:
            if self._message_matches_criteria(msg, criteria):
                self.log.debug('update callback %s with criteria %s',
                               callback, criteria)
                self._loop.call_soon(callback, msg)

    def _parse_product_data_response(self, address, userdata):
        category = userdata[4]
        subcategory = userdata[5]
        firmware = userdata[6]
        self.log.info('INSTEON Product Data Response from %r: cat:%s, subcat:%s',
                      address, hex(category), hex(subcategory))

        self.devices[address.hex] = {'cat': category, 'subcat': subcategory,
                                     'firmware': firmware}

    def _parse_button_event(self, msg):
        self.log.info('PLM button event: %02x (%s)', msg.event, msg.description)

    def _parse_get_plm_info(self, msg):
        self.log.info('PLM Info from %r: category:%02x subcat:%02x firmware:%02x',
                      msg.address, msg.category, msg.subcategory, msg.firmware)
        self._me = msg.address

    def _parse_get_plm_config(self, msg):
        self.log.info('PLM Config: flags:%02x spare:%02x spare:%02x',
                      msg.flagsval, msg.spare1, msg.spare2)

    def _parse_all_link_record(self, msg):
        self.log.info('ALL-Link Record for %r: flags:%02x group:%02x data:%02x/%02x/%02x',
                      msg.address, msg.flagsval, msg.group,
                      msg.linkdata1, msg.linkdata2, msg.linkdata3)
        
        if msg.address.hex in self.devices:
            self.log.info("Device %r is already added manually.", msg.address.hex)
            if msg.address.hex in self._userdefineddevices:
                self._userdefineddevices[msg.address.hex]["status"] = "found"
        else:
            self.log.info("Auto Discovering device %r.", msg.address.hex)
            self.devices[msg.address.hex] = {'cat': msg.linkdata1,
                                         'subcat': msg.linkdata2,
                                         'firmware': msg.linkdata3}

        if self.devices.state == 'loading':
            self.get_next_all_link_record()

    def _parse_all_link_completed(self, msg):
        self.log.info('ALL-Link Completed %r: group:%d cat:%02x subcat:%02x '
                      'firmware:%02x linkcode: %02x',
                      msg.address, msg.group, msg.category, msg.subcategory,
                      msg.firmware, msg.linkcode)

        if msg.address.hex in self.devices:
            self.log.info("Device %r is already added manually.", msg.address.hex)
            if msg.address.hex in self._userdefineddevices:
                self._userdefineddevices[msg.address.hex]["status"] = "found"
        else:
            self.log.info("Auto Discovering device %r.", msg.address.hex)
            self.devices[msg.address.hex] = {'cat': msg.category,
                                         'subcat': msg.subcategory,
                                         'firmware': msg.firmware}
        for userdevice in self._userdefineddevices:
            if self._userdefineddevices[userdevice]["status"] == "notfound":
                self.log.info("Finished all link, and failed to discover device %r.", userdevice)

    def _queue_hex(self, message, wait_for=None):
        if wait_for is None:
            wait_for = {}

        self.log.debug('Adding command to queue: %s', message)
        self._send_queue.append([message, wait_for])

    def _send_hex(self, message, wait_for=None):
        if wait_for is None:
            wait_for = {}

        if self._last_command or self._wait_for:
            self.log.debug('Still waiting on last_command.')
            self._queue_hex(message, wait_for)
        else:
            self._send_raw(binascii.unhexlify(message))
            self._schedule_wait(wait_for)

    def _send_raw(self, message):
        self.log.debug('Sending %d byte message: %s',
                      len(message), binascii.hexlify(message))
        self.transport.write(message)
        self._last_command = message

    def add_message_callback(self, callback, criteria):
        """Register a callback for when a matching message is seen."""
        self._message_callbacks.append([callback, criteria])
        self.log.debug('Added message callback to %s on %s', callback, criteria)

    def add_insteon_callback(self, callback, criteria):
        """Register a callback for when a matching INSTEON command is seen."""
        self._insteon_callbacks.append([callback, criteria])
        self.log.debug('Added INSTEON callback to %s on %s', callback, criteria)

    def add_update_callback(self, callback, criteria):
        """Register as callback for when a matching device attribute changes."""
        self._update_callbacks.append([callback, criteria])
        self.log.debug('Added update callback to %s on %s', callback, criteria)

    def add_device_callback(self, callback, criteria):
        """Register a callback for when a matching new device is seen."""
        self.devices.add_device_callback(callback, criteria)

    def send_insteon_standard(self, device, cmd1, cmd2, wait_for=None):
        """Send an INSTEON Standard message to the PLM."""
        if wait_for is None:
            wait_for = {}

        device = Address(device)
        rawstr = '0262'+device.hex+'00'+cmd1+cmd2
        self._send_hex(rawstr, wait_for)

    # pylint: disable=too-many-arguments
    def send_insteon_extended(self, device, cmd1, cmd2,
                              userdata='0000000000000000000000000000',
                              wait_for=None):
        """Send an INSTEON Extended message to the PLM."""
        if wait_for is None:
            wait_for = {}

        device = Address(device)
        rawstr = '0262'+device.hex+'10'+cmd1+cmd2+userdata
        self._send_hex(rawstr, wait_for)

    def get_plm_info(self):
        """Request PLM Info."""
        self.log.info('Requesting PLM Info')
        self._send_hex('0260')

    def get_plm_config(self):
        """Request PLM Config."""
        self.log.info('Requesting PLM Config')
        self._send_hex('0273')

    def get_first_all_link_record(self):
        """Request first ALL-Link record."""
        self.log.info('Requesting First ALL-Link Record')
        self._send_hex('0269', wait_for={'code': 0x57})

    def get_next_all_link_record(self):
        """Request next ALL-Link record."""
        self.log.info('Requesting Next ALL-Link Record')
        self._send_hex('026a', wait_for={'code': 0x57})

    def load_all_link_database(self):
        """Load the ALL-Link Database into object."""
        self.devices.state = 'loading'
        self.get_first_all_link_record()

    def product_data_request(self, addr):
        """Request Product Data Record for device."""
        device = Address(addr)
        self.log.info('Requesting product data for %s', device.human)
        self.send_insteon_standard(
            device, '03', '00',
            wait_for={'code': 0x51, 'cmd1': 0x03, 'cmd2': 0x00})

    def text_string_request(self, addr):
        """Request Device Text String."""
        device = Address(addr)
        self.log.info('Requesting text string for %s', device.human)
        self.send_insteon_standard(
            device, '03', '02',
            wait_for={'code': 0x51, 'cmd1': 0x03, 'cmd2': 0x02})

    def status_request(self, addr, cmd2='00'):
        """Request Device Status."""
        address = Address(addr)
        device = self.devices[address.hex]

        if 'no_requests' in device.get('capabilities'):
            self.log.debug('Skipping status_request for no_requests device %r (%s)',
                           address, device.get('model', 'Unknown Model'))
            return

        self.log.info('Requesting status for %r', address)

        callback = self._parse_status_response
        if device.get('model') == '2450':
            if cmd2 == '00':
                self._loop.call_later(1, self.status_request, address, '01')
            else:
                callback = self._parse_sensor_response

        elif 'binary_sensor' in device.get('capabilities'):
            callback = self._parse_sensor_response

        self.send_insteon_standard(
            address, '19', cmd2,
            wait_for={'code': 0x50, '_callback': callback})

    def extended_status_request(self, addr):
        """Request Operating Flags for device."""
        device = Address(addr)
        self.log.info('Requesting extended status for %s', device.human)
        self.send_insteon_extended(
            device, '2e', '00',
            wait_for={'code': 0x51, '_callback': self._parse_extended_status_response})

    def update_setlevel(self, addr, level):
        """Currently non-functional."""
        device = Address(addr)
        self.log.info('Changing setlevel on %s to %02x', device.human, level)
        self.send_insteon_extended(
            device, '2e', '00',
            userdata='00067f0000000000000000000000',
            wait_for={'code': 0x50})

    def update_ramprate(self, addr, level):
        """Currently non-functional."""
        device = Address(addr)
        self.log.info('Changing setlevel on %s to %02x', device.human, level)
        self.send_insteon_extended(
            device, '2e', '00',
            userdata='00051b0000000000000000000000',
            wait_for={'code': 0x50})

    def get_device_attr(self, addr, attr):
        """Return attribute on specified device."""
        return self.devices.getattr(addr, attr)

    def turn_off(self, addr):
        """Send command to device to turn off."""
        address = Address(addr)
        self.send_insteon_standard(address, '13', '00', {})

    def turn_on(self, addr, brightness=255, ramprate=None):
        """Send command to device to turn on."""
        address = Address(addr)
        device = self.devices[address.hex]
        self.log.debug('turn_on %r %s', addr, device.get('model'))

        if isinstance(ramprate, int):
            #
            # The specs say this should work, but I couldn't get my 2477D
            # switches to respond.  Leaving the code in place for future
            # hacking.  If you try to use ramprate it probably won't work.
            #
            bhex = 'fc'
            self.send_insteon_standard(address, '2e', bhex, {})
        else:
            bhex = str.format('{:02X}', int(brightness)).lower()
            self.send_insteon_standard(address, '11', bhex, {})

        if device.get('model') == '2450':
            #
            # Request status after two seconds so we can detect if the I/OLinc
            # is configured in 'momentary contact' mode and turned off right
            # after we sent the on.  We can't rely on regular INSTEON state
            # broadcasts with this device because the state broadcasts come
            # from the sensor and not the relay.
            #
            self._loop.call_later(2, self.status_request, address)

    def poll_devices(self):
        """Walk through ALDB and populate device information for each device."""
        self.log.info('Polling all devices in ALDB')
        for address in self.devices:
            self.status_request(address)

    def list_devices(self):
        """Debugging command to expose ALDB."""
        for address in self.devices:
            device = self.devices[address]
            print(address, ':', device)



"""Insteon Powerline Modem Interface Module.

This module provides a unified asyncio network handler for interacting with
Insteon Powerline modems like the 2413U and 2412S.
"""
from .connection import Connection      # noqa: F401
from .protocol import PLM               # noqa: F401



"""Embodies the INSTEON Product Database static data and access methods."""
import logging
import collections

# pylint: disable=line-too-long
# pylint: disable=too-few-public-methods

Product = collections.namedtuple('Product', 'cat subcat product_key description model capabilities')


class IPDB(object):
    """Embodies the INSTEON Product Database static data and access methods."""

    products = [
        Product(0x01, 0x00, None, 'LampLinc 3-pin', '2456D3', ['light', 'dimmer']),
        Product(0x01, 0x01, None, 'SwitchLinc Dimmer (600W)', '2476D', ['light', 'dimmer']),
        Product(0x01, 0x02, None, 'In-LineLinc Dimmer', '2475D', ['light', 'dimmer']),
        Product(0x01, 0x03, None, 'Icon Switch Dimmer', '2476D', ['light', 'dimmer']),
        Product(0x01, 0x04, None, 'SwitchLinc Dimmer (1000W)', '2476DH', ['light', 'dimmable']),
        Product(0x01, 0x06, None, 'LampLinc 2-pin', '2456D2', ['light', 'dimmer']),
        Product(0x01, 0x07, None, 'LampLinc Dimmer V2 2-pin', '2856D2', ['light', 'dimmable']),
        Product(0x01, 0x0d, None, 'SocketLinc Dimmer', '2454D', ['light', 'dimmable']),
        Product(0x01, 0x0e, None, 'LampLinc Dual Band 2-pin', '2457D2', ['light', 'dimmable']),
        Product(0x01, 0x13, 0x000032, 'SwitchLinc Dimmer (Lixar)', '2676D-B', ['light', 'dimmer']),
        Product(0x01, 0x17, None, 'ToggleLinc Dimmer', '2466D', ['light', 'dimmer']),
        Product(0x01, 0x18, 0x00003F, 'Icon SwitchLinc Dimmer Inline Companiion', '2474D', ['light', 'dimmer']),
        Product(0x01, 0x1a, 0x00004F, 'In-LineLinc Dimmer', '2475DA1', ['light', 'dimmable']),
        Product(0x01, 0x1b, 0x000050, 'KeypadLinc Dimmer, 6-button', '2486DWH6', ['light', 'dimmable']),
        Product(0x01, 0x1b, 0x000051, 'KeypadLinc Dimmer, 8-button', '2486DWH8', ['light', 'dimmable']),
        Product(0x01, 0x1d, None, 'SwitchLinc Dimmer (1200W)', '2476D', ['light', 'dimmable']),
        Product(0x01, 0x1e, None, 'Icon Switch Dimmer i2', '2476DB', ['light', 'dimmer']),
        Product(0x01, 0x1f, None, 'ToggleLinc Dimmer', '2466D', ['light', 'dimmer']),
        Product(0x01, 0x20, None, 'SwitchLinc Dimmer (600W)', '2477D', ['light', 'dimmable']),
        Product(0x01, 0x21, None, 'OutletLinc Dimmer', '2472D', ['light', 'dimmable']),
        Product(0x01, 0x22, None, 'LampLinc 2-pin', '2457D2X', ['light', 'dimmer']),
        Product(0x01, 0x23, None, 'LampLinc EZ', '2457D2', ['light', 'dimmer']),
        Product(0x01, 0x24, None, 'SwitchLinc 2-wire Dimmer', '2474D', ['light', 'dimmer']),
        Product(0x01, 0x25, None, 'Ballast Dimmer', '2475DA2', ['light', 'dimmable']),
        Product(0x01, 0x2e, None, 'FanLinc Dual Band', '2475F', ['light', 'dimmable']),
        Product(0x01, 0x31, None, 'SwitchLinc Dimmer 240V', '2478D', ['light', 'dimmer']),
        Product(0x01, 0x32, None, 'In-LineLinc Dimmer', '2457D2', ['light', 'dimmable']),
        Product(0x01, 0x34, None, 'DIN Rail Dimmer', '2542-222', ['light', 'dimmable']),
        Product(0x01, 0x35, None, 'Micro Dimmer', '2442-222', ['light', 'dimmable']),
        Product(0x01, 0x3a, None, 'LED Bulb', '2672-222', ['light', 'dimmable']),
        Product(0x01, 0x41, None, 'KeypadLinc Dimmer, 8 button', '2334-222', ['light', 'dimmable']),
        Product(0x01, 0x42, None, 'KeypadLinc Dimmer, 6 button', '2334-232', ['light', 'dimmable']),
        Product(0x01, 0x49, None, 'Recessed LED Bulb', '2674-222', ['light', 'dimmable']),

        Product(0x02, 0x06, None, 'ApplianceLinc Outdoor 3-pin', '2456S3E', ['switch']),
        Product(0x02, 0x08, None, 'OutletLinc', '2473S', ['switch']),
        Product(0x02, 0x09, None, 'ApplianceLinc 3-pin', '2456S3', ['switch']),
        Product(0x02, 0x10, None, 'In-LineLinc Relay', '2475S', ['switch']),
        Product(0x02, 0x12, None, 'Icon SwitchLinc Relay Inline Companion', '2474S', ['switch']),
        Product(0x02, 0x13, None, 'Icon SwitchLinc Relay (Lixar)', '2676R-B', ['switch']),
        Product(0x02, 0x14, None, 'In-LineLinc Relay with Sense', '2475S2', ['switch']),
        Product(0x02, 0x0a, None, 'SwitchLinc Relay', '2476S', ['light']),
        Product(0x02, 0x0b, None, 'Icon Relay On-Off', '2876SB', ['light']),
        Product(0x02, 0x0c, None, 'Icon Appliance Module 3-pin', '2856SB', ['switch']),
        Product(0x02, 0x0d, None, 'ToggleLinc Relay', '2466S', ['light']),
        Product(0x02, 0x0e, None, 'SwitchLinc Relay Timer', '2476ST', ['switch']),
        Product(0x02, 0x16, None, 'Icon Relay On-Off', '2876SB', ['light']),
        Product(0x02, 0x17, None, 'Icon Appliance Module', '2856SB', ['switch']),
        Product(0x02, 0x1a, None, 'ToggleLinc Relay', '2466SW', ['switch']),
        Product(0x02, 0x1f, None, 'In-LineLinc Relay', '2475SDB', ['switch']),
        Product(0x02, 0x2a, None, 'SwitchLinc Switch', '2477S', ['light']),
        Product(0x02, 0x2e, None, 'DIN Rail On/Off', '2453-222', ['switch']),
        Product(0x02, 0x2f, None, 'Micro On/Off', '2443-222', ['switch']),
        Product(0x02, 0x37, None, 'On/Off Module', '2635-222', ['switch']),
        Product(0x02, 0x38, None, 'On/Off Outdoor Module', '2634-222', ['switch']),
        Product(0x02, 0x39, None, 'On/Off Outlet', '2663-222', ['switch']),

        Product(0x05, 0x0b, None, 'Thermostat', '2441TH', ['climate']), #<- Coming Soon!

        Product(0x07, 0x00, None, 'I/O Linc', '2450', ['switch', 'binary_sensor', 'relay']),

        Product(0x09, 0x0a, None, '220/240V 30A Load Controller NO', '2477SA1', ['switch']),
        Product(0x09, 0x0b, None, '220/240V 30A Load Controller NC', '2477SA2', ['switch']),

        Product(0x10, 0x01, None, 'Motion Sensor', '2842-222', ['binary_sensor']),
        Product(0x10, 0x02, None, 'TriggerLinc', '2421', ['binary_sensor']),
        Product(0x10, 0x08, None, 'Water Leak Sensor', '2852-222', ['binary_sensor']),
        Product(0x10, 0x11, None, 'Hidden Door Sensor', '2845-222', ['binary_sensor', 'no_requests']),

        Product(0x0f, 0x06, None, 'MorningLinc', '2458A1', ['switch']),
    ]

    def __init__(self):
        self.log = logging.getLogger(__name__)

    def __len__(self):
        return len(self.products)

    def __iter__(self):
        for product in self.products:
            yield product

    def __getitem__(self, key):
        cat, subcat = key

        for product in self.products:
            if cat == product.cat and subcat == product.subcat:
                return product

        # We failed to find a device in the database, so we will make a best
        # guess from the cat/subcat.
        #
        name = 'Unknown Device'
        capabilities = []

        if cat == 0x01:
            name = 'Unknown Dimmer'
            capabilities.append('light')
            capabilities.append('dimmer')

        if cat == 0x02:
            name = 'Unknown Device'
            capabilities.append('switch')

        if cat == 0x10:
            name = 'Unknown Sensor'
            capabilities.append('binary_sensor')

        return Product(cat, subcat, None, name, None, capabilities)
